// Neon++ Library, (C) 2006 M. Derezynski
// A C++ wrapper library for neon <http://www.webdav.org/neon>, (C) Joe Orton

#ifndef _NEON_PP_REQUEST_
#define _NEON_PP_REQUEST_

#include <string>
#include <vector>
#include <map>
#include <iostream>
#include <istream>
#include <ostream>

#include <ne_session.h>
#include <ne_request.h>
#include <sigc++/sigc++.h>

#include "session.hh"

namespace Neon
{
  typedef unsigned int HttpPort;

  class Request
  {
    public:

      /** The method to use for a request. Currently GET and POST are supported.
       *
       */
      enum Method
      {
        METHOD_POST,
        METHOD_GET
      };

      /** Whether to read the response from the request, or discard it
       *
       */
      enum ResponseRead
      {
        RESPONSE_READ,
        RESPONSE_IGNORE
      };

      /** Enumeration to use with the Neon::Request::add_request_header() variant that takes in a
       *  HeaderType
       *
       *  HEADER_WWWFORM: Adds Content-type: application/x-www-form-urlencoded to the headers
       *
       */
      enum HeaderType
      {
        HEADER_WWWFORM,
        HEADER_TEXTXML,
      };

#include "neonpp-exception.hh"
      
      NEONPPEXCEPTION(InvalidRequestError)

#ifdef NEONPP_EXCEPTIONS_ENABLED
      NEONPPEXCEPTION(NeAuthError)
      NEONPPEXCEPTION(NeProxyAuthError)
      NEONPPEXCEPTION(NeConnectError)
      NEONPPEXCEPTION(NeTimeoutError)
      NEONPPEXCEPTION(NeGenericError)

      NEONPPEXCEPTION(NoCurrentRequestError)
#endif

      typedef std::map<std::string, std::string> ResponseHeaders;

      /** Creates a request object
       *
       * @param hostname The hostname to use. Note that this shall not contain the preceding http:// or https://
       * @param path The path to get from the host, defaults to '/'
       * @param port The port to use, defaults to port 80
       * @param read Whether to read the response from this request. Defaults to Neon::Request::RESPONSE_READ
       * @param method The method to use for the request. Defaults to Neon::Request::METHOD_GET
       * @param scheme The scheme to use. Can be one of "http" or "https"
       *
       */
      Request (std::string const& hostname,
               std::string const& path = "/", 
               Neon::HttpPort port = 80,
               ResponseRead read = RESPONSE_READ,
               Method method = METHOD_GET,
               std::string const& scheme = "http")

      throw (Neon::Session::SessionCreateError, InvalidRequestError);

      ~Request ();      

      /** Gets the session for a Neon::Request
       *
       * @result Returns the Neon::Session used by the Request
       */
      Neon::Session & get_session ();

      /** Puts a request body in case of a POST request
       *
       * @param data The data to be pushed. Thsi is const char * to accomodate for binary data that might contain zeros in between
       * @param data_size The size of the data to be pushed out
       *
       */
      void set_request_data (const char * data, size_t data_size);


      /** Whether to automatically handle a HTTP 302 (Redirect)
       *
       * @param handle Whether to automatically handle 302 (defaults to true)
       *
       */
      void handle_redirect (bool handle = true);

      /** Adds a request header giving the name and value of the header, e.g. "Content-type" and "text/xml"
       *
       * @param name The name of the header
       * @param value The content of this header
       *
       */
      void add_request_header (std::string const& name, std::string const& value);

      /** Adds a request header using one of the predefined types in Neon::Request::HeaderType
       *
       * @param type Neon::Request::HeaderType
       *
       */
      void add_request_header (HeaderType type);

#ifdef NEONPP_EXCEPTIONS_ENABLED
      /** Dispatches the request manually. Note that this is usually not needed if you use one of the operators << and >>,
       *  as they will call dispatch internally themselves
       *
       */
      void dispatch ();
#else
      /** Dispatches the request manually. Note that this is usually not needed if you use one of the operators << and >>,
       *  as they will call dispatch internally themselves
       *
       * @result The result status as in ne_request_dispatch()
       */
      int dispatch ();
#endif

      /** Resets the request, that is, clears the (potentially read) response data, and allows for re-fetching using
       *  Neon::Request::dispatch(), or operator << or >>
       *
       */
      void clear ();

      /** Gets the ne_status of the current request
       *
       * @result The const (libneon-owned) ne_status struct
       */
      const ne_status * get_status () const;

      /** Gets the current error string, if any
       *
       * @result Error string, if any
       */
      std::string get_error () const;

      /** Gets the response headers 
       *
       * @result the response headers as a map, @see Neon::Request::ResponseHeaders
       */
      void get_response_headers (Neon::Request::ResponseHeaders & headers);

      /** Gets the read data
       *
       * @result The read data as a single data block (owned by the Neon::Session object and thus const)
       */
      const unsigned char * get_data ();

      /** Gets the data size
       *
       * @result The size of the data block
       */
      size_t get_data_size ();

      /** Allows for easily outputting the read context to an std::string
       *  NOTE: This appends a zero at the end to make sure the string is zero-terminated
       *  operator<< as well as get_data() do not modify the data in any way
       *
       */
      void operator>> (std::string&);

      /** Allows for appending the data content into an ostream, like an ofstream or a stringstream
       *
       */
      std::ostream& operator<<(std::ostream& os);

      /** Allows for appending the data content into an ostream, like an ofstream or a stringstream
       *
       */
      friend std::ostream& operator<<(std::ostream& os, Request & r);

    private:

      Request ()
          : m_request (0),
            m_session (0) {}

      static int read_block (void * udata, const char * data, size_t len);

      ne_request   *m_request;
      Session      *m_session;
      bool          m_dispatched;
      ResponseRead  m_read; 
      Method        m_method;      
      bool          m_handle_redirect;
      std::vector<unsigned char> m_vec;
  };
}
#endif
